#include <cstdio>
#include <vector>

using namespace std;

#include "o65linker.hh"
#include "msginsert.hh"
#include "dataarea.hh"
#include "space.hh"

#include <getopt.h>

void MessageLinkingModules(unsigned count)
{
    //printf("Linking %u modules\n", count);
}

void MessageLoadingItem(const string& header)
{
    //printf("Loading %s\n", header.c_str());
}

void MessageWorking()
{
}

void MessageDone()
{
    //printf("- done\n");
}

void MessageModuleWithoutAddress(const string& name)
{
    fprintf(stderr, "O65 linker: Module %s is still without address\n", name.c_str());
}

void MessageUndefinedSymbol(const string& name)
{
    fprintf(stderr, "O65 linker: Symbol '%s' still undefined\n", name.c_str());
}

void MessageDuplicateDefinition(const string& name, unsigned nmods, unsigned ndefs)
{
    fprintf(stderr, "O65 linker: Symbol '%s' defined in %u module(s) and %u global(s)\n",
        name.c_str(), nmods, ndefs);
}

void MessageUndefinedSymbols(unsigned n)
{
    fprintf(stderr, "O65 linker: Still %u undefined symbol(s)\n", n);
}


int main(int argc, char** argv)
{
    std::vector<std::string> files;

    std::FILE *output = NULL;
    std::string outfn;

    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0,0,'h'},
            {"version",  0,0,'V'},
            {"output",   0,0,'o'},
            {0,0,0,0}
        };
        int c = getopt_long(argc,argv, "hVo:", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'V': // version
            {
                printf(
                    "%s %s\n"
                    "Copyright (C) 1992,2005 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "This is free software; see the source for copying conditions. There is NO\n" 
                    "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
                    argv[0], VERSION
                      );
                return 0;
            }
            
            case 'h':
            {
                std::printf(   
                    "65816 linker\n"
                    "Copyright (C) 1992,2005 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: %s [<option> [<...>]] <file> [<...>]\n"
                    "\nLinks O65 and IPS files together and produces an IPS file.\n"
                    "\nOptions:\n"
                    " --help, -h            This help\n"
                    " --version, -V         Displays version information\n"
                    " -o <file>             Places the output into <file>\n"
                    "\nNo warranty whatsoever.\n",
                    argv[0]);
                return 0;
            }

            case 'o':
            {
                outfn = optarg;
                if(outfn != "-")
                {
                    if(output)
                    {
                        std::fclose(output);
                    }
                    output = std::fopen(outfn.c_str(), "wb");
                    if(!output)
                    {
                        std::perror(outfn.c_str());
                        goto ErrorExit;
                    }
                }
                break;
            }
        }
    }

    while(optind < argc)
        files.push_back(argv[optind++]);

    if(files.empty())
    {
        fprintf(stderr, "Error: Link what? See %s --help\n", argv[0]);
    ErrorExit:
        if(output)
        {
            fclose(output);
            remove(outfn.c_str());
        }
        return -1;
    }
    
    
    O65linker linker;
    freespacemap freespace;
    
    // Assume everything is free space!
    /* FIXME: Make this configurable. */
    for(unsigned page=0x00; page<=0x3F; ++page)
        freespace.Add(page, 0x0000, 0x10000);
    
    freespace.Del(0x00, 0x0000, 0x8000); // Not usable
    
    for(unsigned a=0; a<files.size(); ++a)
    {
        char Buf[5];
        FILE *fp = fopen(files[a].c_str(), "rb");
        if(!fp)
        {
            perror(files[a].c_str());
            continue;
        }
        
        fread(Buf, 1, 5, fp);
        if(!strncmp(Buf, "PATCH", 5))
        {
            linker.LoadIPSfile(fp, files[a]);
        }
        else
        {
            O65 tmp;
            tmp.Load(fp);
            
            const vector<pair<unsigned char, string> >&
                customheaders = tmp.GetCustomHeaders();
            
            LinkageWish Linkage;
            
            for(unsigned b=0; b<customheaders.size(); ++b)
            {
                unsigned char type = customheaders[b].first;
                const string& data = customheaders[b].second;
                switch(type)
                {
                    case 10: // linkage type
                    {
                        unsigned param = 0;
                        if(data.size() >= 5)
                        {
                            param = (data[1] & 0xFF) 
                                  | ((data[2] & 0xFF) << 8)
                                  | ((data[3] & 0xFF) << 16)
                                  | ((data[4] & 0xFF) << 24);
                        }
                        switch(data[0])
                        {
                            case 0:
                                Linkage = LinkageWish();
                                break;
                            case 1:
                                Linkage.SetLinkageGroup(param);
                                fprintf(stderr, "%s will be linked in group %u\n",
                                    files[a].c_str(), param);
                                break;
                            case 2:
                                Linkage.SetLinkagePage(param);
                                fprintf(stderr, "%s will be linked to page $%02X\n",
                                    files[a].c_str(), param);
                                break;
                        }
                        break;
                    }
                    case 0: // filename
                    case 1: // operating system header
                    case 2: // assembler name
                    case 3: // author
                    case 4: // creation date
                        break;
                }
            }
            
            linker.AddObject(tmp, files[a], Linkage);
        }
        fclose(fp);
    }
    
    /* Organize the code blobs */    
    freespace.OrganizeO65linker(linker);

    linker.Link();
    
    vector<unsigned> tmp;
    
    std::FILE* stream = output ? output : stdout;

    fprintf(stream, "PATCH");
    
    vector<unsigned> o65addrs = linker.GetAddrList();
    for(unsigned a=0; a<o65addrs.size(); ++a)
    {
        const vector<unsigned char>& code = linker.GetCode(a);
        
        unsigned char data[5];
        
        data[0] = (o65addrs[a] >>16) & 0x3F;
        data[1] = (o65addrs[a] >> 8) & 0xFF;
        data[2] = (o65addrs[a]     ) & 0xFF;
        data[3] = (code.size() >> 8) & 0xFF;
        data[4] = (code.size()     ) & 0xFF;
        
        fwrite(data, 1, 5, stream);
        fwrite(&code[0], 1, code.size(), stream);
    }

    fprintf(stream, "EOF");
    
    if(output) fclose(output);
    
    return 0;
}
